/* ics_version.h */

#pragma once

#define ICS_VERSION_STR "2026.3.3"
#define ICS_FILEVERSION 2026, 3, 3, 0
#define ICS_NAME "ICSx64.dll"
#define ICS_DESCRIPTION "Image Correlation Spectroscopy Library"
#define ICS_AUTHOR "Christoph Gohlke"
#define ICS_COPYRIGHT "Copyright (C) 2016-2026"
#define ICS_COMPANY "cgohlke.com"
